unit rhlKeccak;

interface

uses
  rhlCore;

type

  { TrhlKeccak }

  TrhlKeccak = class abstract(TrhlHash)
  private
    m_state: array[0..24] of QWord;
  protected
    procedure UpdateBlock(const AData); override;
  public
    constructor Create; override;
    procedure Init; override;
    procedure Final(var ADigest); override;
  end;

  { TrhlKeccak224 }

  TrhlKeccak224 = class(TrhlKeccak)
    constructor Create; override;
  end;

  { TrhlKeccak256 }

  TrhlKeccak256 = class(TrhlKeccak)
    constructor Create; override;
  end;

  { TrhlKeccak384 }

  TrhlKeccak384 = class(TrhlKeccak)
    constructor Create; override;
  end;

  { TrhlKeccak512 }

  TrhlKeccak512 = class(TrhlKeccak)
    constructor Create; override;
  end;

implementation

{ TrhlKeccak512 }

constructor TrhlKeccak512.Create;
begin
  HashSize := 512 div 8;
  inherited Create;
end;

{ TrhlKeccak384 }

constructor TrhlKeccak384.Create;
begin
  HashSize := 384 div 8;
  inherited Create;
end;

{ TrhlKeccak256 }

constructor TrhlKeccak256.Create;
begin
  HashSize := 256 div 8;
  inherited Create;
end;

{ TrhlKeccak224 }

constructor TrhlKeccak224.Create;
begin
  HashSize := 224 div 8;
  inherited Create;
end;

{ TrhlKeccak }

procedure TrhlKeccak.UpdateBlock(const AData);
const
  FC: array[0..23] of QWord = (
    $0000000000000001,$0000000000008082,$800000000000808a,$8000000080008000,
    $000000000000808b,$0000000080000001,$8000000080008081,$8000000000008009,
    $000000000000008a,$0000000000000088,$0000000080008009,$000000008000000a,
    $000000008000808b,$800000000000008b,$8000000000008089,$8000000000008003,
    $8000000000008002,$8000000000000080,$000000000000800a,$800000008000000a,
    $8000000080008081,$8000000000008080,$0000000080000001,$8000000080008008);
var
  data: array[0..0] of QWord absolute AData;
  j: Integer;

  Aba, Abe, Abi, Abo, Abu,
  Aga, Age, Agi, Ago, Agu,
  Aka, Ake, Aki, Ako, Aku,
  Ama, Ame, Ami, Amo, Amu,
  Asa, Ase, Asi, Aso, Asu,
  Bba, Bbe, Bbi, Bbo, Bbu,
  Bga, Bge, Bgi, Bgo, Bgu,
  Bka, Bke, Bki, Bko, Bku,
  Bma, Bme, Bmi, Bmo, Bmu,
  Bsa, Bse, Bsi, Bso, Bsu,
  Ca, Ce, Ci, Co, Cu,
  Da, De, Di, Doo, Du,
  Eba, Ebe, Ebi, Ebo, Ebu,
  Ega, Ege, Egi, Ego, Egu,
  Eka, Eke, Eki, Eko, Eku,
  Ema, Eme, Emi, Emo, Emu,
  Esa, Ese, Esi, Eso, Esu: QWord;
begin
  for j := 0 to (BlockSize div 8) - 1 do
    m_state[j] := m_state[j] xor data[j];

  Aba := m_state[0];
  Abe := m_state[1];
  Abi := m_state[2];
  Abo := m_state[3];
  Abu := m_state[4];
  Aga := m_state[5];
  Age := m_state[6];
  Agi := m_state[7];
  Ago := m_state[8];
  Agu := m_state[9];
  Aka := m_state[10];
  Ake := m_state[11];
  Aki := m_state[12];
  Ako := m_state[13];
  Aku := m_state[14];
  Ama := m_state[15];
  Ame := m_state[16];
  Ami := m_state[17];
  Amo := m_state[18];
  Amu := m_state[19];
  Asa := m_state[20];
  Ase := m_state[21];
  Asi := m_state[22];
  Aso := m_state[23];
  Asu := m_state[24];

  Ca := Aba xor Aga xor Aka xor Ama xor Asa;
  Ce := Abe xor Age xor Ake xor Ame xor Ase;
  Ci := Abi xor Agi xor Aki xor Ami xor Asi;
  Co := Abo xor Ago xor Ako xor Amo xor Aso;
  Cu := Abu xor Agu xor Aku xor Amu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor FC[0];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  Fc[1];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[2];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[3];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[4];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[5];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[6];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[7];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[8];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[9];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[10];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[11];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[12];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[13];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[14];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[15];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[16];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[17];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[18];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[19];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[20];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[21];
  Ca := Aba;
  Abe := Bbe xor (not Bbi or Bbo);
  Ce := Abe;
  Abi := Bbi xor (Bbo and Bbu);
  Ci := Abi;
  Abo := Bbo xor (Bbu or Bba);
  Co := Abo;
  Abu := Bbu xor (Bba and Bbe);
  Cu := Abu;
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Ca := Ca xor Aga;
  Age := Bge xor (Bgi and Bgo);
  Ce := Ce xor Age;
  Agi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Agi;
  Ago := Bgo xor (Bgu or Bga);
  Co := Co xor Ago;
  Agu := Bgu xor (Bga and Bge);
  Cu := Cu xor Agu;
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ca := Ca xor Aka;
  Ake := Bke xor (Bki and Bko);
  Ce := Ce xor Ake;
  Aki := Bki xor (not Bko and Bku);
  Ci := Ci xor Aki;
  Ako := (not Bko) xor (Bku or Bka);
  Co := Co xor Ako;
  Aku := Bku xor (Bka and Bke);
  Cu := Cu xor Aku;
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ca := Ca xor Ama;
  Ame := Bme xor (Bmi or Bmo);
  Ce := Ce xor Ame;
  Ami := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Ami;
  Amo := not Bmo xor (Bmu and Bma);
  Co := Co xor Amo;
  Amu := Bmu xor (Bma or Bme);
  Cu := Cu xor Amu;
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Asa;
  Ase := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ase;
  Asi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Asi;
  Aso := Bso xor (Bsu or Bsa);
  Co := Co xor Aso;
  Asu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Asu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Aba := Aba xor  Da;
  Bba := Aba;
  Age := Age xor De;
  Bbe := (Age shl 44) xor (Age shr (64 - 44));
  Aki := Aki xor Di;
  Bbi := (Aki shl 43) xor (Aki shr (64 - 43));
  Amo := Amo xor Doo;
  Bbo := (Amo shl 21) xor (Amo shr (64 - 21));
  Asu := Asu xor Du;
  Bbu := (Asu shl 14) xor (Asu shr (64 - 14));
  Eba := Bba xor (Bbe or Bbi);
  Eba := Eba xor fc[22];
  Ca := Eba;
  Ebe := Bbe xor (not Bbi or Bbo);
  Ce := Ebe;
  Ebi := Bbi xor (Bbo and Bbu);
  Ci := Ebi;
  Ebo := Bbo xor (Bbu or Bba);
  Co := Ebo;
  Ebu := Bbu xor (Bba and Bbe);
  Cu := Ebu;
  Abo := Abo xor Doo;
  Bga := (Abo shl 28) xor (Abo shr (64 - 28));
  Agu := Agu xor Du;
  Bge := (Agu shl 20) xor (Agu shr (64 - 20));
  Aka := Aka xor Da;
  Bgi := (Aka shl 3) xor (Aka shr (64 - 3));
  Ame := Ame xor De;
  Bgo := (Ame shl 45) xor (Ame shr (64 - 45));
  Asi := Asi xor Di;
  Bgu := (Asi shl 61) xor (Asi shr (64 - 61));
  Ega := Bga xor (Bge or Bgi);
  Ca := Ca xor Ega;
  Ege := Bge xor (Bgi and Bgo);
  Ce := Ce xor Ege;
  Egi := Bgi xor (Bgo or not Bgu);
  Ci := Ci xor Egi;
  Ego := Bgo xor (Bgu or Bga);
  Co := Co xor Ego;
  Egu := Bgu xor (Bga and Bge);
  Cu := Cu xor Egu;
  Abe := Abe xor De;
  Bka := (Abe shl 1) xor (Abe shr (64 - 1));
  Agi := Agi xor Di;
  Bke := (Agi shl 6) xor (Agi shr (64 - 6));
  Ako := Ako xor Doo;
  Bki := (Ako shl 25) xor (Ako shr (64 - 25));
  Amu := Amu xor Du;
  Bko := (Amu shl 8) xor (Amu shr (64 - 8));
  Asa := Asa xor Da;
  Bku := (Asa shl 18) xor (Asa shr (64 - 18));
  Eka := Bka xor (Bke or Bki);
  Ca := Ca xor Eka;
  Eke := Bke xor (Bki and Bko);
  Ce := Ce xor Eke;
  Eki := Bki xor (not Bko and Bku);
  Ci := Ci xor Eki;
  Eko := (not Bko) xor (Bku or Bka);
  Co := Co xor Eko;
  Eku := Bku xor (Bka and Bke);
  Cu := Cu xor Eku;
  Abu := Abu xor Du;
  Bma := (Abu shl 27) xor (Abu shr (64 - 27));
  Aga := Aga xor Da;
  Bme := (Aga shl 36) xor (Aga shr (64 - 36));
  Ake := Ake xor De;
  Bmi := (Ake shl 10) xor (Ake shr (64 - 10));
  Ami := Ami xor Di;
  Bmo := (Ami shl 15) xor (Ami shr (64 - 15));
  Aso := Aso xor Doo;
  Bmu := (Aso shl 56) xor (Aso shr (64 - 56));
  Ema := Bma xor (Bme and Bmi);
  Ca := Ca xor Ema;
  Eme := Bme xor (Bmi or Bmo);
  Ce := Ce xor Eme;
  Emi := Bmi xor (not Bmo or Bmu);
  Ci := Ci xor Emi;
  Emo := not Bmo xor (Bmu and Bma);
  Co := Co xor Emo;
  Emu := Bmu xor (Bma or Bme);
  Cu := Cu xor Emu;
  Abi := Abi xor Di;
  Bsa := (Abi shl 62) xor (Abi shr (64 - 62));
  Ago := Ago xor Doo;
  Bse := (Ago shl 55) xor (Ago shr (64 - 55));
  Aku := Aku xor Du;
  Bsi := (Aku shl 39) xor (Aku shr (64 - 39));
  Ama := Ama xor Da;
  Bso := (Ama shl 41) xor (Ama shr (64 - 41));
  Ase := Ase xor De;
  Bsu := (Ase shl 2) xor (Ase shr (64 - 2));
  Esa := Bsa xor (not Bse and Bsi);
  Ca := Ca xor Esa;
  Ese := not Bse xor (Bsi or Bso);
  Ce := Ce xor Ese;
  Esi := Bsi xor (Bso and Bsu);
  Ci := Ci xor Esi;
  Eso := Bso xor (Bsu or Bsa);
  Co := Co xor Eso;
  Esu := Bsu xor (Bsa and Bse);
  Cu := Cu xor Esu;
  Da := Cu xor (Ce shl 1) xor (Ce shr (64 - 1));
  De := Ca xor (Ci shl 1) xor (Ci shr (64 - 1));
  Di := Ce xor (Co shl 1) xor (Co shr (64 - 1));
  Doo := Ci xor (Cu shl 1) xor (Cu shr (64 - 1));
  Du := Co xor (Ca shl 1) xor (Ca shr (64 - 1));
  Eba := Eba xor Da;
  Bba := Eba;
  Ege := Ege xor De;
  Bbe := (Ege shl 44) xor (Ege shr (64 - 44));
  Eki := Eki xor Di;
  Bbi := (Eki shl 43) xor (Eki shr (64 - 43));
  Emo := Emo xor Doo;
  Bbo := (Emo shl 21) xor (Emo shr (64 - 21));
  Esu := Esu xor Du;
  Bbu := (Esu shl 14) xor (Esu shr (64 - 14));
  Aba := Bba xor (Bbe or Bbi);
  Aba := Aba xor  fc[23];
  Abe := Bbe xor (not Bbi or Bbo);
  Abi := Bbi xor (Bbo and Bbu);
  Abo := Bbo xor (Bbu or Bba);
  Abu := Bbu xor (Bba and Bbe);
  Ebo := Ebo xor Doo;
  Bga := (Ebo shl 28) xor (Ebo shr (64 - 28));
  Egu := Egu xor Du;
  Bge := (Egu shl 20) xor (Egu shr (64 - 20));
  Eka := Eka xor Da;
  Bgi := (Eka shl 3) xor (Eka shr (64 - 3));
  Eme := Eme xor De;
  Bgo := (Eme shl 45) xor (Eme shr (64 - 45));
  Esi := Esi xor Di;
  Bgu := (Esi shl 61) xor (Esi shr (64 - 61));
  Aga := Bga xor (Bge or Bgi);
  Age := Bge xor (Bgi and Bgo);
  Agi := Bgi xor (Bgo or not Bgu);
  Ago := Bgo xor (Bgu or Bga);
  Agu := Bgu xor (Bga and Bge);
  Ebe := Ebe xor De;
  Bka := (Ebe shl 1) xor (Ebe shr (64 - 1));
  Egi := Egi xor Di;
  Bke := (Egi shl 6) xor (Egi shr (64 - 6));
  Eko := Eko xor Doo;
  Bki := (Eko shl 25) xor (Eko shr (64 - 25));
  Emu := Emu xor Du;
  Bko := (Emu shl 8) xor (Emu shr (64 - 8));
  Esa := Esa xor Da;
  Bku := (Esa shl 18) xor (Esa shr (64 - 18));
  Aka := Bka xor (Bke or Bki);
  Ake := Bke xor (Bki and Bko);
  Aki := Bki xor (not Bko and Bku);
  Ako := not Bko xor (Bku or Bka);
  Aku := Bku xor (Bka and Bke);
  Ebu := Ebu xor Du;
  Bma := (Ebu shl 27) xor (Ebu shr (64 - 27));
  Ega := Ega xor Da;
  Bme := (Ega shl 36) xor (Ega shr (64 - 36));
  Eke := Eke xor De;
  Bmi := (Eke shl 10) xor (Eke shr (64 - 10));
  Emi := Emi xor Di;
  Bmo := (Emi shl 15) xor (Emi shr (64 - 15));
  Eso := Eso xor Doo;
  Bmu := (Eso shl 56) xor (Eso shr (64 - 56));
  Ama := Bma xor (Bme and Bmi);
  Ame := Bme xor (Bmi or Bmo);
  Ami := Bmi xor (not Bmo or Bmu);
  Amo := not Bmo xor (Bmu and Bma);
  Amu := Bmu xor (Bma or Bme);
  Ebi := Ebi xor Di;
  Bsa := (Ebi shl 62) xor (Ebi shr (64 - 62));
  Ego := Ego xor Doo;
  Bse := (Ego shl 55) xor (Ego shr (64 - 55));
  Eku := Eku xor Du;
  Bsi := (Eku shl 39) xor (Eku shr (64 - 39));
  Ema := Ema xor Da;
  Bso := (Ema shl 41) xor (Ema shr (64 - 41));
  Ese := Ese xor De;
  Bsu := (Ese shl 2) xor (Ese shr (64 - 2));
  Asa := Bsa xor (not Bse and Bsi);
  Ase := not Bse xor (Bsi or Bso);
  Asi := Bsi xor (Bso and Bsu);
  Aso := Bso xor (Bsu or Bsa);
  Asu := Bsu xor (Bsa and Bse);

  m_state[0] := Aba;
  m_state[1] := Abe;
  m_state[2] := Abi;
  m_state[3] := Abo;
  m_state[4] := Abu;
  m_state[5] := Aga;
  m_state[6] := Age;
  m_state[7] := Agi;
  m_state[8] := Ago;
  m_state[9] := Agu;
  m_state[10] := Aka;
  m_state[11] := Ake;
  m_state[12] := Aki;
  m_state[13] := Ako;
  m_state[14] := Aku;
  m_state[15] := Ama;
  m_state[16] := Ame;
  m_state[17] := Ami;
  m_state[18] := Amo;
  m_state[19] := Amu;
  m_state[20] := Asa;
  m_state[21] := Ase;
  m_state[22] := Asi;
  m_state[23] := Aso;
  m_state[24] := Asu;
end;

constructor TrhlKeccak.Create;
begin
  BlockSize := 200 - (HashSize * 2);
end;

procedure TrhlKeccak.Init;
begin
  inherited Init;
  FillChar(m_state, SizeOf(m_state), 0);
  m_state[1] := $FFFFFFFFFFFFFFFF;
  m_state[2] := $FFFFFFFFFFFFFFFF;
  m_state[8] := $FFFFFFFFFFFFFFFF;
  m_state[12] := $FFFFFFFFFFFFFFFF;
  m_state[17] := $FFFFFFFFFFFFFFFF;
  m_state[20] := $FFFFFFFFFFFFFFFF;
end;

procedure TrhlKeccak.Final(var ADigest);
var
  buffer_pos: Integer;
  block: TBytes;
begin
  buffer_pos := FBuffer.Pos;

  FBuffer.SetZeroPadded;
  block := FBuffer.Bytes;

  block[buffer_pos] := 1;
  block[BlockSize - 1] := block[BlockSize - 1] or 128;

  UpdateBlock(block[0]);

  m_state[1] := not m_state[1];
  m_state[2] := not m_state[2];
  m_state[8] := not m_state[8];
  m_state[12] := not m_state[12];
  m_state[17] := not m_state[17];

  Move(m_state[0], ADigest, HashSize);
end;

end.
